/* TIN_BHV.C : behaviour functions */
/* TIN ( Totally Ignorant Nitwits). Copyright (C) 1997 Dirk Jetten */

#include "tin.h"                                /* common stuff */
#include "tin_glob.h"                           /* global stuff */

/* -------------------------------------------------------------- */
static int trace_present( int i)
/* positive when there's a scent trace in the cell where tin[ i]
 * resides and he/she not in direct surrounding
 */
{
     register tin_t  *tp = tin[ i];
     register cell_t *ep = env[ tp->trail[ 0].row][ tp->trail[ 0].col];
     pos_t            dummy;

     if ((ep->scnt_trail >  0) &&                      /* smelling someone */
	     (ep->scnt_tin   != i) &&                                /* not me */
	     (ep->scnt_tin   != ENV_FREE)) {
             /* if it resides in an adjacent cell return negative */
             if ( env_near_occupied( ep->scnt_tin,
                                     tp->trail[ 0],
                                     &dummy.row, &dummy.col)) {
	            dbgprint( "Debug: trace_present #0 end of quest. %d at (%d,%d)\n",
                          ep->scnt_tin, dummy.row, dummy.col);
	            return( -tp->heritage.imp_follow_scnt);
             }
             else {
	            dbgprint( "Debug: trace_present #1 smelling %d\n",
                          ep->scnt_tin );
	            return( tp->heritage.imp_follow_scnt);
             }
     }
     else {
	     return( -tp->heritage.imp_follow_scnt);
     }
}
     
/* -------------------------------------------------------------- */
static int how_hungry( int i, int neighbr)
/* positive when there's food around in cell or luggage or
 * weaker neighbour, and we're running low on fat
 */
{
     register tin_t      *tp = tin[ i];
     register tin_t      *np = tin[ neighbr];
     register cell_t     *ep = env[ tp->trail[ 0].row][ tp->trail[ 0].col];

     if ((( ep->food != 0) ||
	      ( tp->luggage.food != 0)) &&
	     ( tp->fatrsv < FATR_LOW)) {
	     return( tp->heritage.imp_eat);
     }
     else {
#if ACT_CANNIBAL
         if (( np != NULL) && (np->alive) &&
	         (utl_surv_score( i) > utl_surv_score( neighbr))) {
	         return( tp->heritage.imp_eat);
	     }
#endif
	     return( -tp->heritage.imp_eat);
     }
}

/* -------------------------------------------------------------- */
static int how_horny( int i, int neighbr)
/* positive when reproduction seems possible
 * Entry condition : neighbour has got to exist
 */
{
     register tin_t *tp = tin[ i];      
     register tin_t *np = tin[ neighbr];    

     /* basic check on success */
     if (( tp->sex          != np->sex)    &&
	     (!tp->pregnant)  && (!np->pregnant) &&

#ifdef MAXCHLD
        ( tp->children_had  <= MAXCHLD)     &&
        ( np->children_had  <= MAXCHLD)     &&
#endif

	    ( tp->age           > tp->heritage.age_adult)   &&
	    ( np->age           > np->heritage.age_adult)) {
	      return( tp->heritage.imp_mate);
     }
     else {
	      return( -tp->heritage.imp_mate);
     }
}

/* -------------------------------------------------------------- */
static bool family( int i, int j)
/* determines if 2 tin's are related in 1st degree by comparing heritance */
{
     register tin_t *tp = tin[ i];      
     register tin_t *op = tin[ j];   
     register bool  tmp = FALSE;

	 if ((tp->heritage.father_id == op->heritage.father_id) ||
		 (tp->heritage.mother_id == op->heritage.mother_id) ||
	     (tp->heritage.father_id == op->id)    ||
	     (tp->heritage.mother_id == op->id)) {
	     tmp = TRUE;
     }
	 dbgprint( "Debug: family #0 family = %d\n", tmp);
     return( tmp);
}

/* -------------------------------------------------------------- */
static int how_mad( int i, int j)
/* no agression against members of opposite sex
 * Entry condition : neighbour has got to exist
 */
{
     register tin_t      *tp = tin[ i];
     register tin_t      *np = tin[ j];

#if KILLMALL
     /* no conscience whatsoever */
     if (1) {
#else
     /* no agression against related in 1st degree neighbours */
     if ( !family( i, j) &&
          (tp->sex == np->sex)) {
#endif
	    return( tp->heritage.imp_agression);
     }
     else {
	    return( -tp->heritage.imp_agression);
     }
}

/*
--------------------------------------------------------------
*/

static int how_greedy( int i)
{
     register tin_t      *tp = tin[ i];
     register cell_t     *ep = env[ tp->trail[ 0].row][ tp->trail[ 0].col];

     if ((ep->food != 0) &&
	 (tp->luggage.food < tp->heritage.gathered_enough)) {
          return( tp->heritage.imp_gather_food);
     }
     else {
	      return( -tp->heritage.imp_gather_food);
     }
}

/*
--------------------------------------------------------------
*/

behav_t bhv_intention( int i, int *neighbr, char *info)
/* what to do next ?
 * This routine is kept simple&stupid; the more cunning the
 * algorithm, the less decided by selection.
 */
{
     register tin_t *tp = tin[ i];
     cell_t         *ep = env[ tp->trail[ 0].row][ tp->trail[ 0].col];

     /* following values are computed to simulate the drive for
      * certain behaviour. the higher, the more eager to act accordingly */
     int            mate_val   = -MAXINT;
     int            hunger_val = -MAXINT;
     int            gather_val = -MAXINT;
     int            follow_val = -MAXINT;
     int            fight_val  = -MAXINT;
     int            move_val   = -MAXINT;

     behav_t        idea    = nsf;                 /* nsf = nothing so far */
     int            max_val = tp->heritage.imp_relax;            /* sync ^ */

     /* life sucks, and then you die. */
     if (tp->fatrsv <= FATR_DIE) {
	  strcpy( info, " of hunger");
	  return( die);
     }
     if (tp->age >= tp->heritage.age_die) {
	  strcpy( info, " of age");
	  return( die);
     }

     /* neighbour present ? if so, any impulse triggered by this event ? */
     if ((*neighbr = env_neighbour( i)) != -1) {
      dbgprint( "Debug: bhv_intention #0 neighbour %d at (%d,%d)\n",
                *neighbr, tin[ *neighbr]->trail[ 0].row,
                          tin[ *neighbr]->trail[ 0].col);
	  mate_val   = how_horny( i, *neighbr);
	  fight_val  = how_mad( i, *neighbr);
     }
     else {
          dbgprint( "Debug: bhv_intention #1 no neighbour spotted\n");
     }

     /* hunger is a special case, which can involve presence of neighbour */
     hunger_val = how_hungry( i, *neighbr);

     /* check drives that do not need neighbour presence */
     follow_val = trace_present( i);
     move_val   = tp->heritage.imp_move;
     gather_val = how_greedy( i);

     dbgprint( "Debug: bhv_intention #2 mate=%d, fight=%d, eat=%d, follow=%d, move=%d, gather=%d\n",
	    mate_val, fight_val, hunger_val, follow_val, move_val, gather_val);


     if ( mate_val > max_val) {
	  idea = mate;
	  max_val  = mate_val;  
     }    
     if ( hunger_val > max_val) {
	  idea = eat;
	  max_val  = hunger_val;  
     }    
     if ( follow_val > max_val) {
	  idea = follow_scnt;
	  max_val  = follow_val;  
     }    
     if ( move_val > max_val) {
	  idea = move;
	  max_val = move_val;
     }    
     if ( gather_val > max_val) {
	  idea = gather_food;
     }    
     if ( fight_val > max_val) {
	  max_val = fight_val;
	  if (utl_surv_score( i) > utl_surv_score( *neighbr)) {
	       idea    = kill;
	       sprintf( info, "killed by %d", i);
	    }
	    else {
	       idea    = die;
	       sprintf( info, "killed by %d", *neighbr);
	    }
     }
     return( idea);
}
