/* TIN_ENV.C : environment related functions */
/* TIN ( Totally Ignorant Nitwits). Copyright (C) 1997 Dirk Jetten */

#include "tin.h"
#include "tin_glob.h"                           /* global decl. */

/* -------------------------------------------------------------- */

int env_neighbour( int i)
/* returns tin index of first found neighbour of tin i, -1 if not found.
 * maybe a next version will include an array of found neighbours.
 */
{
     register  int j = 0;
     bool      searching = TRUE;
     int       myrow = tin[ i]->trail[ 0].row;
     int       mycol = tin[ i]->trail[ 0].col;

     while (( tin[ j] != NULL) && searching ){
          /* next expr works only with left to right shortcut evaluation */
          if ( (tin[ j]->alive) && (j != i) &&
               (abs( tin[ j]->trail[ 0].row - myrow) < 2) &&
               (abs( tin[ j]->trail[ 0].col - mycol) < 2)) {
               searching = FALSE;
          }
          else {
               dbgprint("Debug: env_neighbour #0 it is not %d at (%d,%d)\n", j,
                                       tin[ j]->trail[ 0].row,
                                       tin[ j]->trail[ 0].col);
               dbgprint("Debug: env_neighbour #1 because dx = %d, dy = %d\n",
                                       abs( tin[ j]->trail[ 0].row - myrow),
                                       abs( tin[ j]->trail[ 0].col - mycol));
               j++;
          }
     }
     return( searching ? -1 : j);
}


/* -------------------------------------------------------------- */
bool env_free( pos_t position)
/* check if environment cell contains no tin's or other thingies */
{
    int i = 0;
    register pr  = position.row;
    register pc  = position.col;
    register cell_t *ep = env[ pr][ pc];

    if (( pr > 0) && ( pr < env_x_max-1) &&
        ( pc > 0) && ( pc < env_y_max-1) &&
	( ep != NULL)) {
        return ((ep->tin_no == ENV_FREE) ? TRUE : FALSE);
    }
    else {
        return( FALSE);
    }
}
/* -------------------------------------------------------------- */

bool env_near_free( pos_t base, int *row, int *col)
/* try to find a free neighbouring environment cell to position 'base' 
 * the resulting row and column are returned.
 */
{
     bool        gotcha = FALSE;         /* temporary result */
     headng_t    look = nw;
     pos_t       newbase;

     dbgprint( "Debug: env_near_free #0 base (row,col): (%d,%d)\n",
                base.row, base.col);

     *row = ENV_WALL;
     *col = ENV_WALL;

     while (!gotcha && look <= n) {   
          utl_position( base, look, &newbase.row, &newbase.col);
          if (gotcha = env_free( newbase)){
               *row = newbase.row;
               *col = newbase.col; 
               dbgprint("Debug: env_near_free #1 (%d,%d) is free\n",
                                newbase.row, newbase.col);
          }
          look++;
     }
     return( gotcha);  
}

/* -------------------------------------------------------------- */

bool env_near_occupied( int i, pos_t base, int *row, int *col)
/* checks if an adjacent cell is occupied by tin[ i] */
{
     bool        gotcha = FALSE;         /* temporary result */
     headng_t    look = nw;
     pos_t       newbase;
     cell_t      *ep;

     dbgprint( "Debug: env_near_occupied #0 base (row,col): (%d,%d)\n",
                base.row, base.col);

     *row = ENV_WALL;
     *col = ENV_WALL;

     while (!gotcha && look <= n) {   
          utl_position( base, look, &newbase.row, &newbase.col);
          if (env[ newbase.row][ newbase.col]->tin_no == i){
               *row   = newbase.row;
               *col   = newbase.col;
               gotcha = TRUE;
               dbgprint("Debug: env_near_occupied #1 found %d at (%d,%d)\n",
                                i, newbase.row, newbase.col);
          }
          look++;
     }
     return( gotcha);  
}


/* -------------------------------------------------------------- */


bool env_create( void)
/* creates the environment cells and food sources */
{
    int i, j;

    /* create the cells */
    for (i=0; i < env_x_max; i++){
          for (j=0; j < env_y_max; j++){
               env[ i][ j] = (cell_t *) malloc( sizeof( cell_t));
 
               if ( env[ i][ j] == NULL){
                    perror("Error: allocating environment memory");
                    exit( 1);
               }
               else{
                    env[ i][ j] = memset( env[i][j], ENV_FREE, sizeof( cell_t));
		            if (env[ i][ j] == NULL) {
                        perror("Error: memset environment cell");
                        exit( 1);
		            }
                    else {
              	        dbgprint( "Debug: env_create #0 created envcell (%d,%d)\n",
                                  i, j);
			            env[ i][ j]->food = 0;
		            }
               }
          }
     }

     /* initialize the cells */
     /* walls around the border */
     for ( i=0; i < env_x_max; i++){
          env[ i][           0]->tin_no = ENV_WALL;   
          env[ i][ env_y_max-1]->tin_no = ENV_WALL;
     }
     for ( j=0; j < env_y_max; j++){
          env[ 0][           j]->tin_no = ENV_WALL;   
          env[ env_x_max-1][ j]->tin_no = ENV_WALL;
     }

     /* create food suppliers */
     for ( i = 0; i < turnip_max; i++) {
          turnip[ i] = NULL;
     }

     for ( i = 0; i < no_turnips; i++) {
          turnip[ i] = (food_t *)malloc( sizeof( food_t));
	      if (turnip[ i] == NULL){
	          perror("Error: allocating turnip memory");
              exit( 1);
	      }
	      else {
              turnip[ i]->position.row = utl_random( env_x_max-1);
              turnip[ i]->position.col = utl_random( env_y_max-1);
              dbgprint( "Debug: env_create #1 created turnip at (%d,%d)\n",
			  turnip[ i]->position.row, turnip[ i]->position.col);
	      }
     }
     return( TRUE);
}

/* -------------------------------------------------------------- */
void env_grow( int i) 
/* the cycle of food supplier [ i] */
{
     cell_t    *ep = env[ turnip[ i]->position.row][ turnip[i]->position.col];

     if ( ep->food < MAXINT) {
        ep->food++;
     }
}

/* -------------------------------------------------------------- */
void env_trail_update( void) {

	 int       row, col;

     /* age scent trail in environment */
     /* ..can this be done more efficient ? */
     for (row =1; row < env_x_max; row++) {
         for (col =1; col < env_y_max; col++) {
             if (env[ row][ col]->scnt_trail > 0) {
                 env[ row][ col]->scnt_trail--;
             }
         }
     }
}
