/* TIN_MAIN.C */
/* TIN ( Totally Ignorant Nitwits). Copyright (C) 1997 Dirk Jetten */

#include "tin.h"                                /* common stuff */
#include "tin_glob.h"                           /* global decl. */
#include "getopt.h"                             /* GNU getopt */

/* default heritage initialization */
int     init_imp_eat         = 4000;
int     init_imp_mate        = 6000;
int     init_imp_move        = 110;
int     init_imp_relax       = 105;
int     init_imp_follow      = 3500;
int     init_imp_agression   = 100;
int     init_imp_gather_food = 4000;

int     init_age_die          = 1800;
int     init_age_adult        = 0;
int     init_chance_male      = 500;
int     init_eaten_enough     = 100;
int     init_gathered_enough  = 50;
int     init_fatrsv           = 1000;

/* -------------------------------------------------------------- */
static bool init_tin_create( )
/* initial creation of tins */
{
     register    int i;
     tin_t       *newborn = NULL;
     int         home;
     bool        ok = TRUE;

     if (ok) { 
          newborn = utl_tin_create( &home, init_chance_male);
          ok = ( newborn != NULL);
     }

     if (ok) {
         herit_t *nhp = &newborn->heritage;

         dbgprint( "Debug: init_tin_create #0 newborn %d\n", newborn);
         newborn->generation   = 0;
         newborn->fatrsv       = init_fatrsv;
         newborn->luggage.food = 10;
         nhp->age_adult        = utl_mutate( init_age_adult, glob_mutate);
         if (nhp->age_adult < 1 ) nhp->age_adult = 1;
         nhp->age_die          = utl_mutate( init_age_die, glob_mutate);
         if (nhp->age_die < 1 ) nhp->age_die = 1;
         nhp->eaten_enough     = utl_mutate( init_eaten_enough, glob_mutate);
         nhp->gathered_enough = utl_mutate( init_gathered_enough,glob_mutate);
         nhp->imp_eat          = utl_mutate( init_imp_eat,      glob_mutate);
         nhp->imp_mate         = utl_mutate( init_imp_mate,     glob_mutate);
         nhp->imp_agression    = utl_mutate( init_imp_agression,glob_mutate);
         nhp->imp_gather_food  = utl_mutate( init_imp_gather_food,glob_mutate);
         nhp->imp_follow_scnt  = utl_mutate( init_imp_follow,   glob_mutate);
         nhp->imp_move         = utl_mutate( init_imp_move,     glob_mutate);
         nhp->imp_relax        = utl_mutate( init_imp_relax,    glob_mutate);
#if ONEROOT
         /* initial population family in 1st degree */
         nhp->father_id = 1200;
         nhp->mother_id = 1201;
#else
         /* initial population probably not related in 1st degree */
         nhp->father_id = utl_random( MAXINT);
         nhp->mother_id = utl_random( MAXINT);
#endif
         newborn->id = utl_random( MAXINT);
     }
     return( ok);
}

/* -------------------------------------------------------------- */
static void help_on_tin( void)
{
    printf("\tTIN : Totally Ignorant Nitwits engine\n");
    printf("By: Dirk Jetten ( D.Jetten@PTT-Telecom.NL)\n");
    printf("Options :                                         Defaults :\n");
    printf("  -l : logging level\t\t\t\t%s\n",
              loglevels[ glob_log_level]);
    printf("  -m : max mutation percentage\t\t\t\t%f\n",
              glob_mutate);
    printf("  -s : random generator seed\n");
    printf("  -x : environment dimension in # of cells\t\t%d\n",
             env_x_max);
    printf("  -y : environment dimension in # of cells\t\t%d\n",
             env_y_max);
    printf("  -A : initial Agression impulse\t\t\t%d\n",
              init_imp_agression);
    printf("  -D : initial Dying age\t\t\t%d\n",
              init_age_die);
    printf("  -E : initial impulse to Eat\t\t\t\t%d\n",
              init_imp_eat);
    printf("  -F : initial Fat reserve\t\t\t\t%d\n",
              init_fatrsv);
    printf("  -G : initial impulse to Gather food\t\t\t%d\n",
              init_imp_gather_food);
    printf("  -M : initial impulse to Move\t\t\t\t%d\n",
              init_imp_move);
    printf("  -N : initial impulse to do Nothing\t\t\t\t%d\n",
              init_imp_relax);
    printf("  -O : chances on Male 1st generation ( out of 1000)\t%d\n",
              init_chance_male);
    printf("  -Q : Quit eating when fatreserve equals e\t\t%d\n",
              init_eaten_enough);
    printf("  -R : initial impulse to Reproduce\t\t\t%d\n",
              init_imp_mate);
    printf("  -S : initial impulse to follow Scent trails\t\t%d\n",
              init_imp_follow);
    printf("  -T : initial Tin count \t\t\t\t%d\n",
              no_tins);
    printf("  -V : initial Vegetable (turnip) count \t\t%d\n",
              no_turnips);
    printf("  -X : initial age when seX allowed\t\t\t%u\n",
              init_age_adult);
    printf("  -d : enable lifecycle debugging (You Don't Want This)\n");

    printf("\nSee WHATSUP.DOC for detailed info\n");
}
/* -------------------------------------------------------------- */

static void help_on_commands( void)
/* debugging commands */
{
    printf("\nTIN : Totally Ignorant Nitwits\n");
    printf("------------------------------\n\n");
    printf("artificial life (, but not as we know it)\n");
    printf("By: Dirk Jetten\n");
    printf("Options :\n");
    printf(" <number> : number of time-slices to go\n");
    printf(" k        : tin    kill\n");
    printf(" t        : tin    dump\n");
    printf(" e        : cell   dump\n");
    printf(" v        : food   dump\n");
    printf(" f        : foetus dump\n");
    printf(" b        : set/cancel behaviour breakpoint\n");
    printf(" w        : set/cancel behaviour watchpoint\n");
    printf(" o        : output on/off\n");
    printf(" s        : tin statistics\n");
#if 0
    printf(" S        : overall statistics\n");
#endif
    printf(" l        : change logging level (now %s)\n",
              loglevels[ glob_log_level]);
    printf(" d        : low level debug on/off\n");
    printf(" h        : this screen\n");
    printf(" q        : quit\n");
    printf("\n");
    printf("See WHATSUP.DOC for detailed info\n");
}

/* -------------------------------------------------------------- */
static bool analyze_input( u_long *go_cycles, bool *quit,
                           behav_t *bbreak, behav_t *bwatch) {

       bool go = TRUE;
       int  i, row, col;
       char stepstr[25];

       printf("command> ");
       *go_cycles = atoi( gets( stepstr)); printf("\n");
       if (*go_cycles == 0) {                           /* enter = one step */
            *go_cycles = 1;
       }

       if (( strcmp( stepstr, "h") == 0) ||
           ( strcmp( stepstr, "?") == 0)){                          /* help */
            help_on_commands();
            go = FALSE;
       }

       *quit = ( strcmp( stepstr, "q") == 0);                       /* quit */
       if ( strcmp( stepstr, "t") == 0) {                       /* tin dump */
            printf("tin # : ");
            utl_dump_tin( atoi( gets( stepstr)));
            printf("\n");
            go = FALSE;
       }
       if ( strcmp( stepstr, "l") == 0) {                /* change loglevel */
            for ( i = 0; i < 7; i++) {
                printf( "%d = %s,", i, loglevels[ i]);
            }
            printf("\n");
            printf("Current loglevel : %s\n", loglevels[ glob_log_level]);
            printf("Enter loglevel : ");
            glob_log_level = atoi( gets( stepstr));
            printf("\n");
            go = FALSE;
       }
       if ( strcmp( stepstr, "k") == 0) {                       /* tin kill */
            printf("tin # : ");
            if ((tin[(i = atoi( gets( stepstr)))] != NULL)) {
               act_terminate( i, "killed by operator");
            }
            printf("\n");
            go = FALSE;
       }
       if ( strcmp( stepstr, "b") == 0) {                          /* break */
            for ( i = 1; i < 9; i++) {
                printf( "%d = %s,", i, actions[ i]);
            }
            printf("\n");
            if (*bbreak != 0) {
               printf("current breakpoint : %s\n", actions[ *bbreak]);
               printf("enter 0 to cancel\n");
            }
            else {
               printf("no break set\n");
            }
            printf("break # : "); *bbreak = atoi( gets( stepstr));
            if ( *bbreak == 0) {
                printf("\n\tbreak cancelled.\n");
                go = FALSE;
            }
            else {
                printf("\n# Cycles to go : ");
                *go_cycles = atoi( gets( stepstr)); printf("\n");
            }
       }
       if ( strcmp( stepstr, "w") == 0) {                          /* watch */
            for ( i = 1; i < 9; i++) {
                printf( "%d = %s,", i, actions[ i]);
            }
            printf("\n");
            if (*bwatch != 0) {
               printf("current watchpoint : %s\n", actions[ *bwatch]);
               printf("enter 0 to cancel\n");
            }
            else {
               printf("no watchpoint set\n");
            }
            printf("watch # : "); *bwatch = atoi( gets( stepstr));
            printf("\n");
            if ( *bwatch == 0) {
                printf("watch cancelled.\n");
                go = FALSE;
            }
            else {
                printf("# Cycles to go : ");
                *go_cycles = atoi( gets( stepstr));
                printf("\n");
            }
       }
       if ( strcmp( stepstr, "e") == 0) {                  /* env cell dump */
            printf("row    # : ");   row = atoi( gets( stepstr));
            printf("\ncolumn # : "); col = atoi( gets( stepstr));
            printf("\n");
            utl_dump_cell( row, col);
            go = FALSE;
       }
       if ( strcmp( stepstr, "f") == 0) {                    /* foetus dump */
            printf("foetus # : ");   i = atoi( gets( stepstr));
            printf("\n");
            utl_dump_foetus( i);
            go = FALSE;
       }
       if ( strcmp( stepstr, "v") == 0) {                    /* turnip dump */
            printf("turnip # : ");   i = atoi( gets( stepstr));
            printf("\n");
            utl_dump_turnip( i);
            go = FALSE;
       }
       if ( strcmp( stepstr, "o") == 0) {                   /* output switch */
            glob_output_on = !glob_output_on;
	        printf( "output set to %d\n", glob_output_on);
            go = FALSE;
       }
       if ( strcmp( stepstr, "d") == 0) {                   /* debug switch */
            glob_debug = !glob_debug;
	        printf( "debug set to %d\n", glob_debug);
            go = FALSE;
       }
       if ( strcmp( stepstr, "s") == 0) {                 /* tin statistics */
            printf("tin # : ");   i = atoi( gets( stepstr));
            printf("\n");
            utl_dump_tinstats( i);
            go = FALSE;
       }
#if 0
       if ( strcmp( stepstr, "S") == 0) {             /* overall statistics */
            stat_col();
            stat_print();
            go = FALSE;
       }
#endif
       return( go);
}

/* -------------------------------------------------------------- */
static void start( void)
{
     register int i;

     u_long cycles              = 1;
     u_long total_cycles        = 0;
     bool quit                  = FALSE;
     bool go                    = FALSE;
     behav_t behaviour_break    = nsf;
     behav_t behaviour_watch    = nsf;
     behav_t behaviour          = nsf;
     u_int   prev_popul         = 0;
     u_int   max_popul          = 0;
     u_int   min_popul          = MAXINT;

     while (!quit) {
          if (glob_debug || glob_dump) {
             cycles--;
             if (population == 0) {
                go     = FALSE;
                cycles = 0;
             }
             if ( population != prev_popul ) {
                if (population > max_popul) max_popul = population;
                if (population < min_popul) min_popul = population;
                dmpprint( "Info: population now %u, max %u, min %u cycle %lu\n",
                          population, max_popul, min_popul, total_cycles);
                prev_popul = population;
             }
          }
          if (go) {
             total_cycles++;
             for ( i= 0; i < tin_max; i++) {
               if (( tin[ i] != NULL)&&
                   ( tin[ i]->alive == TRUE)) {
                   dbgprint( "Debug: start #0 tin %d cycle\n", i);
                   behaviour = tin_cycle( i);
                   if ((behaviour       == behaviour_break) &&
                       (behaviour_break != nsf            )) {
                       printf("%s", glob_dmpstr);
                       cycles = 0;
                   }
                   if ((behaviour       == behaviour_watch) &&
                       (behaviour_watch != nsf            )) {
                       printf("%s", glob_dmpstr);
                   }
               }
             }
             for ( i= 0; i < turnip_max; i++) {
                 if ( turnip[ i] != NULL) {
                     dbgprint( "Debug: start #1 turnip %d cycle\n", i);
                     env_grow( i);
                 }
             }
             dbgprint( "Debug: start #2 trail update\n");
             env_trail_update();
          }
          strprint("Cycle: population #%d at (0,0) dummy\n", population);
          if ((glob_debug || glob_dump) && (cycles < 1)) {
              printf("Info: env + tin memory = %lu, cycles = %lu\n",
                     ( (u_long) (env_x_max  * env_y_max * sizeof( cell_t))) +
                     ( (u_long) (population * sizeof( tin_t))), total_cycles);
              go = analyze_input( &cycles, &quit,
                                  &behaviour_break, &behaviour_watch);
          }
     }
}
/* -------------------------------------------------------------- */

main(int argc, char **argv)

{
     register int   i,j;
     int            seed;
     int            opt;
     bool           ok = TRUE;

     /* no input check yet */
     while ((opt = getopt(argc, argv, 
                          "d::?hl:m:s:x:y:A:D:E:F:G:M:N:Q:R:S:T:V:X:")) != -1) {
          switch (opt) {
               case 'd': /* debugging */
                    glob_debug = atoi( optarg);
                    break;
               case '?':
               case 'h': /* print options */
                    help_on_tin();
                    return( 0);
               case 'l': /* logging level */
                    glob_log_level = atoi( optarg);
                    break;
               case 'm': /* heritage mutation percentage */
                    glob_mutate = atof( optarg);
                    break;
               case 's': /* seed for random generator */
                    seed = atoi( optarg);
                    srand( seed);
                    break;
               case 'x': /* environment # of rows */
                    if (( env_x_max = atoi( optarg)) > MAXENVX) {
			           printf("Error: max rows = %d\n", MAXENVX);
			           exit( 1);
		            }
                    break;
               case 'y': /* environment # of columns */
                    if (( env_y_max = atoi( optarg)) > MAXENVY) {
			           printf("Error: max columns = %d\n", MAXENVY);
			           exit( 1);
		            }
                    break;
               case 'A': /* initial impulse to fight */
                    init_imp_agression = atoi( optarg);
                    break;
               case 'D': /* initial dying age */
                    init_age_die = atoi( optarg);
                    break;
               case 'E': /* initial impulse to eat */
                    init_imp_eat = atoi( optarg);
                    break;
               case 'F': /* initial fat reserve */
                    init_fatrsv = atoi( optarg);
                    break;
               case 'G': /* initial impulse to gather food */
                    init_imp_gather_food = atoi( optarg);
                    break;
               case 'M': /* initial impulse to move */
                    init_imp_move = atoi( optarg);
                    break;
               case 'N': /* initial impulse to do Nothing */
                    init_imp_relax = atoi( optarg);
                    break;
               case 'O': /* chances on male 1st generation */
                    init_chance_male = atoi( optarg);
                    break;
               case 'Q': /* enough fatreserve built, quit stuffin' */
                    init_eaten_enough = atoi( optarg);
                    break;
               case 'R': /* initial impulse to reproduce */
                    init_imp_mate = atoi( optarg);
                    break;
               case 'S': /* initial impulse to follow scent trails */
                    init_imp_follow = atoi( optarg);
                    break;
               case 'T': /* initial tin population to create */
                    if ((no_tins = atoi( optarg)) > tin_max) {
			           printf("Error: max tins = %d\n", tin_max);
			           exit( 1);
                    }
                    break;
               case 'V': /* initial # of vegetables */
                    if ((no_turnips = atoi( optarg)) > turnip_max) {
			           printf("Error: max turnips = %d\n", turnip_max);
			           exit( 1);
                    }
                    break;
               case 'X': /* initial age when sex allowed */
                    init_age_adult = atoi( optarg);
                    break;
            }
      }

/* init pointers */
      for (i=0; i < tin_max; i++){
         tin[ i] = NULL;
      }

      for (i=0; i < turnip_max; i++){
         turnip[ i] = NULL;
      }

      for (i=0; i < env_x_max; i++){
         for (j=0; j < env_y_max; j++){
             env[ i][ j] = NULL;
          }
      }

      if (ok=env_create( )) {
         for ( i = 0; i < no_tins; i++) {
             if (ok = init_tin_create()) {
                 population++;
                 dbgprint( "Debug: main #0 tin %d created\n", i);
             }
             else {
                 dbgprint( "Debug: main #0 error creating tin %d\n", i);
                 strprint( "Error: error creating tin %d\n", i);
             }
         }
      }

      if (ok) {
         start();
      }
      return( !ok);
}



