/* TIN_UTL.C : utility functions */
/* TIN ( Totally Ignorant Nitwits). Copyright (C) 1997 Dirk Jetten */

#include "tin.h"                                /* common stuff */
#include "tin_glob.h"                           /* global decl. */

/* -------------------------------------------------------------- */
float utl_surv_score( int i) {
/*
 * fitness function, or : who wins in a fight ?
 * factors that influence surviving a fight are :
 * - % female advantage, 1.0 means equally strong.
 * - muscle power
 */

     register tin_t *tp = tin[ i];      
     register float female_adv = ((tp->sex == male) ? 1.0 : glob_female_adv);
     register float tmpf = 0;
     register unsigned tmp = 0;

     if ( tp->age != 0) { /* percentage of life doing physical things */
	    tmpf = (float)( tp->stats.no_move + tp->stats.no_mate +
                        tp->stats.no_kill + tp->stats.no_follow_scnt) /
               (float)( tp->age);
	    dbgprint( "Debug: utl_surv_score #0 muscle = %f\n", tmpf);
	    dbgprint( "Debug: utl_surv_score #1 total  = %f\n",
                  tmpf * female_adv);
     }

	 return( female_adv * tmpf);
}

/* -------------------------------------------------------------- */
int utl_random( int range)
/* generate a random integer number between 1 and <range> */
{
    register int j = 1+(int) (1.0*range*rand()/(RAND_MAX+1.0));
    return (j);
}

/* -------------------------------------------------------------- */
float utl_randomf( float range)
/* generate a random floating point number between 1 and <range> */
{
    float j = 1.0 + (float) (range*rand()/(RAND_MAX+1.0));
    return (j);
}

/* -------------------------------------------------------------- */
void utl_position( pos_t start, headng_t heading, int *row, int *col)
/* returns position (row,column) with given base position and direction */
{
    register int tr  = start.row;
    register int tc  = start.col;

    dbgprint( "Debug: utl_position #0 args : (%d,%d), %s\n", 
              tr, tc, direction[ heading]);
    if ( heading == nw ){   
       tr--; tc++;
    }
    if ( heading == w){   
       tr--;
    }
    if ( heading == sw){
       tr--; tc--;
    }
    if ( heading == s){   
       tc--;
    }
    if ( heading == se){   
       tr++; tc--;
    }
    if ( heading == e){   
       tr++;
    }
    if ( heading == ne){   
       tr++; tc++;
    }
    if ( heading == n){   
       tc++;
    }

    *row = tr;
    *col = tc;
    dbgprint( "Debug: utl_position #1 result: (%d,%d)\n", 
              *row, *col);
}

/* -------------------------------------------------------------- */
int utl_get_free_tinslot( void)
/* get a free tin index, -1 if not found */
{
     int         i;

     dbgprint( "Debug: utl_get_free_tin_slot #0 tin_max = %d\n", tin_max);
     for ( i = 0; i < tin_max; i++){
          if ((tin[ i] == NULL) ||
              ((tin[ i] != NULL) && (!tin[ i]->alive)))break;
     }
     if ( i == tin_max) i = -1;
     dbgprint( "Debug: utl_get_free_tin_slot #1 result: %d\n", i);
     return( i);
}

/* -------------------------------------------------------------- */
tin_t *utl_tin_create( int *home, int male_chances)
/* create the initial tin structure, find a random birthplace */
{
     int         tries = 0; 
     int         i, j;
     int         maxtries = env_x_max * env_y_max * 10; /* 10 is arbitrary*/ 
     bool        ok;
     pos_t       envpos;
     tin_t       *newborn = NULL;

     dbgprint( "Debug: utl_tin_create #0 xmax= %d, ymax=%d\n",
               env_x_max, env_y_max);
     *home = utl_get_free_tinslot(); 
     if (( ok = (*home != -1))) { 
         if (tin[ *home] == NULL) { 
             dbgprint( "Debug: utl_tin_create #1 get memory\n");
             newborn = (tin_t *) malloc( sizeof( tin_t)); 
             if (!(ok = ( newborn != NULL))) { 
                 dbgprint( "Debug: utl_tin_create #2 malloc error\n");
		         perror("Error: allocating tin memory");
		         exit(1);
             }
             else {
                 tin[ *home]        = newborn; 
                 tin[ *home]->alive = FALSE;          /* not really needed */
             }
         } 
         else {
             newborn = tin[ *home];
             tin[ *home]->alive = FALSE;              /* not really needed */
         }
     }
     else {
         strprint( "Error: maximum configured population reached\n");
         dbgprint( "Debug: utl_tin_create #3 too crowded\n");
     }
 
     if (ok){  
         memset( newborn, (char)0, sizeof( tin_t));
         dbgprint( "Debug: utl_tin_create #4 find envcell\n");
         ok = FALSE;
         do { 
             envpos.row = utl_random( env_x_max-1);
             envpos.col = utl_random( env_y_max-1);
             ok = env_free( envpos);
         } 
         while ((!ok) && (++tries <= maxtries)); 
         if (!ok) {
             strprint( "Error: no room for baby\n");
         }
     }
 
     if (ok && (newborn != NULL)){ 
         newborn->alive = TRUE;
         newborn->sex = ( utl_random( 1000) < male_chances ? male : female);
         newborn->pregnant = FALSE;
         newborn->children_had = 0;
         newborn->age   = 0; 
         newborn->prev_bhviour = nsf; 

         newborn->trail[ 0].row = envpos.row;
         newborn->trail[ 0].col = envpos.col;
         env[ envpos.row][ envpos.col]->tin_no = *home;

         newborn->heading = utl_random( 7);
         for ( i = 1; i < MAXTRAIL; i++) {
             newborn->trail[ i].row = ENV_WALL;
             newborn->trail[ i].col = ENV_WALL; 
         } 
         newborn->stats.no_nsf = 0; 
         newborn->stats.no_mate = 0; 
         newborn->stats.no_kill = 0; 
         newborn->stats.no_eat = 0; 
         newborn->stats.no_move = 0; 
         newborn->stats.no_follow_scnt = 0; 
         newborn->stats.no_gather = 0; 
     }
     return (newborn);
}

/* -------------------------------------------------------------- */
static int round( float f)
/*
 * isn't there some ANSI runtime function for this ?
 */
{
    if ( f > 0.0) {
       if ((int) f == (int) (f + 0.5)) {
          return ((int) f);
       }
       else {
            return ((int) (f + 1.0));
       }
    }
    else {
       if ((int) f == (int) (f - 0.5)) {
            return ((int) f);
       }
       else {
            return ((int) (f - 1.0));
       }
    }
}

/* -------------------------------------------------------------- */
int utl_mutate( int value, int mutation)
/* changes given value by a random percentage between
 * minus and plus "mutation" percent
 */
{
   u_int reslt = value;
   float mut_percent = (float)mutation * (float)value / 100.0;
   float tmp;

   dbgprint( "Debug: utl_mutate #0 %d +- %d percent\n",
             value, mutation);
   tmp = utl_randomf( mut_percent);         /* 0 < tmp < mut_percent */
  
   dbgprint( "Debug: utl_mutate #1 +- %f \n", tmp);
   if (utl_random( 2) == 1) {                       /* flips a coin */
       tmp = -tmp;                                    /* tails ! 8-) */
   }

   reslt += round( tmp);
   dbgprint( "Debug: utl_mutate #2 rounded result = %d\n", reslt);

   return( reslt); 
}

/* -------------------------------------------------------------- */
void utl_dump_tin( int i)
/* dumps tin contents to stdout */
{
     tin_t   *tp           = tin[ i];
     u_short j, max, keep;
     herit_t heritage;

    if ( tp == NULL) {
        dmpprint( "Debug: never created\n");
     }
     else {
          if (!tp->alive) {
               dmpprint( "Debug: ....DEAD....\n");
          }
          dmpprint("Debug: id, sex    : %d %s\n", tp->id, gender[ tp->sex]);
          dmpprint("Debug: pregnant   : %d, cycles to go %d, foetus %d\n",
                           tp->pregnant, tp->countdown, tp->foetus);
          dmpprint("Debug: generation : %d\n", tp->generation);
          dmpprint("Debug: heading    : %s\n", direction[ tp->heading]);
          dmpprint("Debug: trail      : (%d,%d) (%d,%d) (%d,%d) (%d,%d) (%d,%d) (%d,%d)\n",
                  tp->trail[ 0].row, tp->trail[ 0].col,
                  tp->trail[ 1].row, tp->trail[ 1].col,
                  tp->trail[ 2].row, tp->trail[ 2].col,
                  tp->trail[ 3].row, tp->trail[ 3].col,
                  tp->trail[ 4].row, tp->trail[ 4].col,
                  tp->trail[ 5].row, tp->trail[ 5].col);
          dmpprint("Debug: age        : %d\n", tp->age);
          dmpprint("Debug: children   : %d\n", tp->children_had);

          dmpprint("Debug: fatrsv     : %d\n", tp->fatrsv);
          dmpprint("Debug: luggage    : %d\n", tp->luggage.food);
          dmpprint("Debug: prev_bhv   : %s\n", actions[ tp->prev_bhviour]);
            dmpprint("Debug: heritage :\n");
          dmpprint("Debug:   max age      : %d, adult age    : %d\n",
                   tp->heritage.age_die, tp->heritage.age_adult);
          dmpprint("Debug:   imp_eat        : %d, imp_mate     : %d\n",
                   tp->heritage.imp_eat, tp->heritage.imp_mate);
          dmpprint("Debug:   imp_agression  : %d, imp_gather   : %d\n",
                   tp->heritage.imp_agression, tp->heritage.imp_gather_food);
          dmpprint("Debug:   imp_follow     : %d, imp_move     : %d\n",
                    tp->heritage.imp_follow_scnt, tp->heritage.imp_move);
          dmpprint("Debug:   imp_relax      : %d\n", tp->heritage.imp_relax);
          dmpprint("Debug:   father_id      : %d\n", tp->heritage.father_id);
          dmpprint("Debug:   mother_id      : %d\n", tp->heritage.mother_id);
          dmpprint("Debug:   stops eating   : %d\n", tp->heritage.eaten_enough);
          dmpprint("Debug:   stops gathering: %d\n", tp->heritage.gathered_enough);
/*

make heritage an array, keep it in sync with actions

          dmpprint("Debug: prior : ");
          heritage = tp->heritage;
          max      = heritage[ 0];
          keep     = 0;
          for ( j = 0, j < 9; j++) {
              for ( i = 0, i < 9; i++) {
                  if (heritage[ i] > max) {
                     max  = heritage [ i];
                     keep = i;
                     heritage [ i] = 0;
                     dmpprint( "%s, ",actions[ i]);
                  }
              }
          }
          dmpprint("\n");




          dmpprint("Debug: stats    :\n");
          dmpprint("Debug:   # nsf         : %d\n", tp->stats.no_nsf);
          dmpprint("Debug:   # mate        : %d\n", tp->stats.no_mate);
          dmpprint("Debug:   # kill        : %d\n", tp->stats.no_kill);
          dmpprint("Debug:   # eat         : %d\n", tp->stats.no_eat);
          dmpprint("Debug:   # move        : %d\n", tp->stats.no_move);
          dmpprint("Debug:   # track       : %d\n", tp->stats.no_follow_scnt);
          dmpprint("Debug:   # gather      : %d\n", tp->stats.no_gather);
*/
          if (!tp->alive) {
                dmpprint( "Debug: ....DEAD....\n");
          }
     }
}

/* -------------------------------------------------------------- */
void utl_dump_tinstats( int i)
/* dumps statistics of individual tins */
{
    tin_t *tp = tin[ i];

    if ( tp == NULL) {
        dmpprint( "Debug: never created\n");
     }
     else {
          if (!tp->alive) {
               dmpprint( "Debug: ....DEAD....\n");
          }

          dmpprint("Debug: stats    :\n");
          dmpprint("Debug:   # nsf         : %d\n", tp->stats.no_nsf);
          dmpprint("Debug:   # mate        : %d\n", tp->stats.no_mate);
          dmpprint("Debug:   # kill        : %d\n", tp->stats.no_kill);
          dmpprint("Debug:   # eat         : %d\n", tp->stats.no_eat);
          dmpprint("Debug:   # move        : %d\n", tp->stats.no_move);
          dmpprint("Debug:   # track       : %d\n", tp->stats.no_follow_scnt);
          dmpprint("Debug:   # gather      : %d\n", tp->stats.no_gather);

          if (!tp->alive) {
                dmpprint( "Debug: ....DEAD....\n");
          }
     }
}

/* -------------------------------------------------------------- */
void utl_dump_cell( int row, int col)
/* dumps cell contents to stdout */
{
    cell_t *ep = env[ row][ col];

    dmpprint("Debug: Dump of environment cell (%d,%d)\n", row, col);
    if ( ep == NULL) {
       dmpprint( "Debug: utl_dump_cell # never created\n");
    }
    else {
          if ( ep->tin_no == ENV_WALL) {
               dmpprint( "Debug: ....WALL....\n");
          }
          else {
               dmpprint("Debug: tin #    : %d\n", ep->tin_no);
               dmpprint("Debug: trail    : %d,\n",ep->scnt_trail);
               dmpprint("Debug: left by  : %d\n", ep->scnt_tin);
               dmpprint("Debug: food eqv : %d\n", ep->food);
        }
    }
}

/* -------------------------------------------------------------- */
void utl_dump_foetus( int i)
/* dumps foetus contents to stdout */
{
     unborn_t *fp = foetus[ i];

    dmpprint("Debug: Dump of foetus %d\n", i);
    if ( fp == NULL) {
       dmpprint( "Debug: utl_dump_foetus never created\n");
    }
    else {
          if ( !fp->occupied) {
               dmpprint( "Debug: ....FREE....\n");
          }
          dmpprint("Debug:   age_die      : %d\n", fp->heritage.age_die);
          dmpprint("Debug:   age_adult    : %d\n", fp->heritage.age_adult);
          dmpprint("Debug:   imp_eat      : %d\n", fp->heritage.imp_eat);
          dmpprint("Debug:   imp_mate     : %d\n", fp->heritage.imp_mate);
          dmpprint("Debug:   imp_agression: %d\n", fp->heritage.imp_agression);
          dmpprint("Debug:   imp_gather   : %d\n",
                    fp->heritage.imp_gather_food);
          dmpprint("Debug:   imp_follow   : %d\n",
                    fp->heritage.imp_follow_scnt);
          dmpprint("Debug:   imp_move     : %d\n", fp->heritage.imp_move);
          dmpprint("Debug:   imp_relax    : %d\n", fp->heritage.imp_relax);
          dmpprint("Debug:   father_value : %d\n", fp->heritage.father_id);
          dmpprint("Debug:   mother_value : %d\n", fp->heritage.mother_id);
          dmpprint("Debug:   stops eating : %d\n",
                    fp->heritage.eaten_enough);
          dmpprint("Debug:   stops gathering: %d\n",
                    fp->heritage.gathered_enough);

    }
}

/* -------------------------------------------------------------- */

void utl_dump_turnip( int i)
/* dumps turnip contents to stdout */
{
    food_t *tp = turnip[ i];
    cell_t *ep;

    dmpprint("Debug: Dump of turnip %d\n", i);
    if ( tp == NULL) {
       dmpprint( "Debug: utl_dump_turnip never created\n");
    }
    else {
       ep = env[ tp->position.row][ tp->position.col];
       dmpprint( "Debug: %d turnips located at (%d,%d)\n",
                 ep->food, tp->position.row, tp->position.col);
    }
}
